/*
   File:           fairplay.c

   Created:        January 4, 2001
   
   Modified:       February 18, 2001

   Author:         Gunnar Andersson (gunnar@nada.kth.se)

   Contents:       A program which cracks messages encrypted with the
                   Playfair cryptosystem.  Shotgun hill-climbing with
		   trigram statistics is used.
*/



#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


#define TRUE                     1
#define FALSE                    0 

#define BUFFER_SIZE              200

#define MINIMUM_TRIGRAM_FREQ     1e-8



typedef struct key {
  int row[26];
  int col[26];
  int grid[8][8];  /* 5x5 used, but 8x8 gives faster addressing */
} Key;



static double start_time;
static double candidate_score;
static double trigram_freq[26][32][32];
static double trigram_score[26624];  /* 26 * 32 * 32 */
static int most_common, second_most_common;
static int grids_tested;
static int *cipher_text;
static int *plain_text;
static int *candidate_text;
static struct key candidate_key;
static struct stat { int count; int index; } cipher_stat[26];



/*
  GET_RANDOM
*/

static int
get_random( void ) {
  return random() / 1024;
}



/*
  GET_CPU_TIME
*/

static double
get_cpu_time( void ) {
  return clock() / (double) CLOCKS_PER_SEC;
}



/*
  GENERATE_KEY
  Generate a grid and the cross reference tables.  The distribution from
  which the grid is taken is skewed such that the most commonly occurring
  letters in the ciphertexts are always put in the same row or column as
  at least one of the two most common letters in the trigram statistics
  file.
*/

static void
generate_key( Key *key ) {
  int i, j;
  int pos;
  int most_common_pos, second_most_common_pos;
  int likely_count;
  int vec[25];
  int likely[25];
  int used[26];

  for ( i = 0; i < 25; i++ )
    vec[i] = -1;
  for ( i = 0; i < 26; i++ )
    used[i] = FALSE;

  /* Place the two most common letters in the statistics randomly
     into the grid. */

  j = get_random() % 25;
  vec[j] = most_common;
  used[most_common] = TRUE;
  most_common_pos = j;

  j = get_random() % 25;
  while ( vec[j] != -1 )
    j = (j + 1) % 25;
  vec[j] = second_most_common;
  used[second_most_common] = TRUE;
  second_most_common_pos = j;

  /* Find the positions in the same row and/or column as at least
     one of the two most common letters. */

  likely_count = 0;
  for ( i = 0; i < 25; i++ )
    if ( ((most_common_pos / 5 == i / 5) ||
	  (most_common_pos % 5 == i % 5) ||
	  (second_most_common_pos / 5 == i / 5) ||
	  (second_most_common_pos % 5 == i % 5)) &&
	 ((most_common_pos != i) && (second_most_common_pos != i)) )
      likely[likely_count++] = i;

  /* Put the most common letters in the cipher-text in the places
     just found. */

  for ( i = 0; i < 5; i++ ) {
    j = get_random() % likely_count;
    vec[likely[j]] = cipher_stat[i].index;
    used[cipher_stat[i].index] = TRUE;
    likely[j] = likely[--likely_count];
  }

  /* Distribute the rest of the alphabet except for J. */

  for ( i = 0; i < 26; i++ )
    if ( !used[i] && (i != 'j' - 'a') ) {
      j = get_random() % 25;
      while ( vec[j] != -1 )
	j = (j + 1) % 25;
      vec[j] = i;
    }

  /* Transfer the grid to the key grid and update cross reference data. */

  for ( i = 0, pos = 0; i < 5; i++ )
    for ( j = 0; j < 5; j++, pos++ ) {
      key->grid[i][j] = vec[pos];
      key->row[vec[pos]] = i;
      key->col[vec[pos]] = j;
    }
}



/*
  SWAP
  Exchanges the letters found in positions (row1,col1) and (row2,col2).
*/

static void
swap( Key *key, int row1, int col1, int row2, int col2 ) {
  int letter1, letter2;

  letter1 = key->grid[row1][col1];
  letter2 = key->grid[row2][col2];

  key->grid[row1][col1] = letter2;
  key->grid[row2][col2] = letter1;

  key->row[letter1] = row2;
  key->col[letter1] = col2;
  key->row[letter2] = row1;
  key->col[letter2] = col1;
}



/*
  TEST_KEY
  Computes the probability for the plaintext using trigram statistics.
*/

static int
test_key( Key *key, int *cipher, int length,
	  double *best_score, double *best_local_score ) {
  double score;
  int i;
  int new_local_optimum;
  int mod5[10] = { 0, 1, 2, 3, 4, 0, 1, 2, 3, 4 };
  unsigned int index;

  i = 0;
  while ( i < length ) {
    int first_row = key->row[cipher[i]];
    int first_col = key->col[cipher[i]];
    int second_row = key->row[cipher[i + 1]];
    int second_col = key->col[cipher[i + 1]];

    if ( first_row == second_row ) {
      plain_text[i] = key->grid[first_row][mod5[first_col + 4]];
      plain_text[i + 1] = key->grid[second_row][mod5[second_col + 4]];
    }
    else if ( first_col == second_col ) {
      plain_text[i] = key->grid[mod5[first_row + 4]][first_col];
      plain_text[i + 1] = key->grid[mod5[second_row + 4]][second_col];
    }
    else {
      plain_text[i] = key->grid[first_row][second_col];
      plain_text[i + 1] = key->grid[second_row][first_col];
    }
    i += 2;
  }

  score = 0.0;
  if ( length >= 3 ) {
    index = (plain_text[0] << 5) + plain_text[1];
    i = 2;
    while ( i < length ) {
      index = ((index << 5) + plain_text[i]) & 32767;
      score += trigram_score[index];
      i++;
    }
  }

  new_local_optimum = FALSE;
  if ( score > *best_local_score + 0.001 ) {
    *best_local_score = score;
    new_local_optimum = TRUE;
    if ( score > *best_score + 0.001 ) {
      *best_score = score;

      candidate_key = *key;
      candidate_score = score;
      for ( i = 0; i < length; i++ )
	candidate_text[i] = plain_text[i];
    }
  }

  return new_local_optimum;
}



/*
  MAIN
*/

int
main( int argc, char *argv[] ) {
  char buffer[BUFFER_SIZE];
  double highest_freq;
  double minimum_score;
  double initial_score, best_local_score, best_score;
  double last_output_score;
  double letter_freq[26];
  int i, j, k, l, m;
  int failures;
  int ch;
  int first_letter, second_letter, third_letter;
  int cipher_length;
  int improved_score;
  int allocation;
  int pair_swap[24][2] = {
    { 0, 2 }, { 1, 2 }, { 2, 3 }, { 0, 3 }, { 1, 2 }, { 2, 3 },
    { 0, 1 }, { 2, 3 }, { 0, 2 }, { 1, 2 }, { 2, 3 }, { 1, 2 },
    { 2, 3 }, { 0, 3 }, { 1, 3 }, { 2, 3 }, { 1, 2 }, { 0, 1 },
    { 2, 3 }, { 1, 2 }, { 0, 1 }, { 2, 3 }, { 1, 2 }, { 0, 1 } };
  Key current_key;
  FILE *in_stream;
  FILE *stat_stream;

  if ( argc != 3 ) {
    fputs( "Usage:\n  fairplay <statistics file> <cipher-text file>\n",
	   stderr );
    exit( EXIT_FAILURE );
  }

  stat_stream = fopen( argv[1], "r" );
  if ( stat_stream == NULL ) {
    fprintf( stderr, "Cannot open statistics file %s\n", argv[1] );
    exit( EXIT_FAILURE );
  }

  for ( i = 0; i < 26; i++ ) {
    letter_freq[i] = 0.0;
    for ( j = 0; j < 26; j++ )
      for ( k = 0; k < 26; k++ )
	trigram_freq[i][j][k] = 0.0;
  }

  fgets( buffer, BUFFER_SIZE, stat_stream );
  while ( !feof( stat_stream ) ) {
    first_letter = toupper( buffer[0] ) - 'A';
    second_letter = toupper( buffer[1] ) - 'A';
    third_letter = toupper( buffer[2] ) - 'A';
    sscanf( buffer, "%*s %lf",
	    &trigram_freq[first_letter][second_letter][third_letter] );
    fgets( buffer, BUFFER_SIZE, stat_stream );
  }

  minimum_score = log( MINIMUM_TRIGRAM_FREQ );
  for ( i = 0; i < 26; i++ )
    for ( j = 0; j < 26; j++ )
      for ( k = 0; k < 26; k++ ) {
	letter_freq[i] += trigram_freq[i][j][k] / 3.0;
	letter_freq[j] += trigram_freq[i][j][k] / 3.0;
	letter_freq[k] += trigram_freq[i][j][k] / 3.0;
	if ( trigram_freq[i][j][k] < MINIMUM_TRIGRAM_FREQ )
	  trigram_score[(32 * i + j) * 32 + k] = minimum_score;
	else
	  trigram_score[(32 * i + j) * 32 + k] = log( trigram_freq[i][j][k] );
      }

  fclose( stat_stream );

  highest_freq = -1.0;
  for ( i = 0; i < 26; i++ )
    if ( letter_freq[i] > highest_freq ) {
      highest_freq = letter_freq[i];
      most_common = i;
    }

  highest_freq = -1.0;
  for ( i = 0; i < 26; i++ )
    if ( (i != most_common) && (letter_freq[i] > highest_freq) ) {
      highest_freq = letter_freq[i];
      second_most_common = i;
    }

  in_stream = fopen( argv[2], "r" );
  if ( in_stream == NULL ) {
    fprintf( stderr, "Cannot open cipher-text file %s\n", argv[2] );
    exit( EXIT_FAILURE );
  }

  allocation = 100;
  cipher_text = malloc( allocation * sizeof( *cipher_text ) );
  plain_text = malloc( allocation * sizeof( *plain_text ) );
  candidate_text = malloc( allocation * sizeof( *candidate_text ) );

  for ( i = 0; i < 26; i++ ) {
    cipher_stat[i].index = i;
    cipher_stat[i].count = 0;
  }
  cipher_length = 0;
  ch = fgetc( in_stream );
  while ( ch != EOF ) {
    ch = tolower( ch );
    if ( islower( ch ) ) {
      if ( ch == 'j' )
	ch = 'i';
      cipher_text[cipher_length++] = ch - 'a';
      cipher_stat[ch - 'a'].count++;
    }
    if ( cipher_length == allocation ) {
      allocation *= 2;
      cipher_text = realloc( cipher_text, 
			     allocation * sizeof( *cipher_text ) );
      plain_text = realloc( plain_text, allocation * sizeof( *plain_text ) );
      candidate_text = realloc( candidate_text,
				allocation * sizeof( *plain_text ) );
    }
    ch = fgetc( in_stream );
  }

  fclose( in_stream );

  for ( i = 0; i < 25; i++ )
    for ( j = 0; j < 25; j++ )
      if ( cipher_stat[j].count < cipher_stat[j + 1].count ) {
	struct stat temp = cipher_stat[j];
	cipher_stat[j] = cipher_stat[j + 1];
	cipher_stat[j + 1] = temp;
      }

  printf( "%d letters in cipher-text\n", cipher_length );

  if ( cipher_length % 2 == 1 ) {
    fputs( "Error:  Odd number of letters in cipher-text\n", stderr );
    exit( EXIT_FAILURE );
  }

  srandom( time( NULL ) );

  start_time = get_cpu_time();

  initial_score = minimum_score * cipher_length - 1.0;
  best_score = initial_score;
  best_local_score = initial_score;
  last_output_score = initial_score;
  generate_key( &current_key );
  test_key( &current_key, cipher_text, cipher_length,
	    &best_score, &best_local_score );
  grids_tested = 1;
  failures = 0;

  while ( 1 ) {
    improved_score = FALSE;

    /* Try to swap characters */

    for ( i = 0; i < 5; i++ )
      for ( j = 0; j < 5; j++ )
	for ( k = i; k < 5; k++ )
	  for ( l = 0; l < 5; l++ )
	    if ( (i != k) || (j != l) ) {
	      swap( &current_key, i, j, k, l );
	      grids_tested++;
	      if ( test_key( &current_key, cipher_text, cipher_length,
			     &best_score, &best_local_score ) ) {
		improved_score = TRUE;
		goto AFTER_LOOPS;
	      }
	      else
		swap( &current_key, i, j, k, l );
	    }

    /* Try all permutations of the rows */

    for ( i = 0; i < 24; i++ ) {
      for ( j = 0; j < 5; j++ )
	swap( &current_key, pair_swap[i][0], j, pair_swap[i][1], j );
      grids_tested++;
      if ( test_key( &current_key, cipher_text, cipher_length,
		     &best_score, &best_local_score ) ) {
	improved_score = TRUE;
	goto AFTER_LOOPS;
      }
    }

    /* Try all permutations of the columns */

    for ( i = 0; i < 24; i++ ) {
      for ( j = 0; j < 5; j++ )
	swap( &current_key, j, pair_swap[i][0], j, pair_swap[i][1] );
      grids_tested++;
      if ( test_key( &current_key, cipher_text, cipher_length,
		     &best_score, &best_local_score ) ) {
	improved_score = TRUE;
	goto AFTER_LOOPS;
      }
    }

  AFTER_LOOPS:

    if ( !improved_score ) {
      /* Can't improve on the current local optimum.  First try perturbations
         of increasing magnitude for a while, then start from a brand new
         grid. */
      failures++;
      if ( failures < 3 )
	for ( i = 0; i < 2 * failures + 4; i++ ) {
	  j = get_random() % 5;
	  k = get_random() % 5;
	  l = get_random() % 5;
	  m = get_random() % 5;
	  if ( (j == l) && (k == m) )
	    m = (m + 1) % 5;
	  swap( &current_key, j, k, l, m );
	}
      else {
	/* If the global optimum has changed since last time we got here,
	   print it on the screen. */
	if ( candidate_score > last_output_score ) {
	  printf( "\n-------------- Score %.3f  ------------------\n\n",
		  candidate_score );
	  for ( i = 0; i < 5; i++ )
	    for ( j = 0; j < 5; j++ ) {
	      printf( " %c", 'a' + candidate_key.grid[i][j] );
	      if ( j == 4 )
		puts( "" );
	    }
	  puts( "" );
	  for ( i = 0; i < cipher_length; i++ ) {
	    putchar( 'a' + candidate_text[i] );
	    if ( (i % 60 == 59) && (i != cipher_length - 1) )
	      puts( "" );
	  }
	  puts( "" );
	  printf( "\n%d grids tested in %.2f seconds\n", grids_tested,
		  get_cpu_time() - start_time );
	}
	last_output_score = candidate_score;

	generate_key( &current_key );
	failures = 0;
      }
      best_local_score = initial_score;
      test_key( &current_key, cipher_text, cipher_length, &best_score,
		&best_local_score );
    }
  }

  return EXIT_SUCCESS;
}
